package zgs.httpd


protected abstract class HOut(aliveReq : Boolean, 
                              resp : HResponse, 
                              chunked : Boolean, 
                              stream : HStream, 
                              session : Option[HSess]) {
  
  //-------- to implement ------------
  
  def write(ar : Array[Byte], offset : Int, length : Int) : Unit
  def close : Unit
  def isHeaded : Boolean
  def isClosed : Boolean

  //----------- common ---------------
  
  protected def closeStream : Unit = if(aliveReq) resp.getHeader("Connection") match {
    case None    => stream.out.switchToRead
    case Some(x) => 
      if(x.toLowerCase.indexOf("close") >= 0) stream.out.close
      else stream.out.switchToRead
  } else stream.out.close
  
  protected def restamp : Unit            = if(session.isDefined) session.get.restamp
  protected def writeSessionCookie : Unit = if(session.isDefined) session.get.writeSessionCookie
}


final protected class HIdentOut(aliveReq : Boolean, 
                                resp : HResponse, 
                                chunked : Boolean, 
                                stream : HStream, 
                                session : Option[HSess]) 
  extends HOut(aliveReq, resp, chunked, stream, session) {
    
  private var headed = false // single thread
  private var closed = false // single thread

  def isHeaded : Boolean = headed
  def isClosed : Boolean = closed
  
  def write(ar : Array[Byte], offset : Int, length : Int) : Unit = {
    
    writeHeader 

    if(chunked) {
      stream.out.write((length.toHexString.toLowerCase + "\r\n").getBytes("ISO-8859-1"))
      stream.out.write(ar, offset, length)
      stream.out.write("\r\n".getBytes("ISO-8859-1"))
    }
    else stream.out.write(ar, offset, length)   
  }
  
  private def writeHeader : Unit = if(!headed) {
    
    if(!aliveReq) resp.setHeader("Connection", "Close") /* else - done in HConn */
    
    if(chunked) {
      resp.removeHeader("Content-Length")
      resp.setHeader("Transfer-Encoding", "chunked")
    }
    else if(resp.getHeader("content-length").isEmpty) {
      stream.dispose
      throw new HError("Content-Length is not set")
    }
    
    writeSessionCookie
    headed = true
    stream.out.write(resp.getBytes)
  }
  
  def close : Unit = if(!closed) {
    closed = true
    writeHeader // for case with empty content
    restamp
    if(chunked) stream.out.write("0\r\n\r\n".getBytes("ISO-8859-1"))
    closeStream
  }
  
}
  

protected class HBufferedOut(aliveReq : Boolean, 
                             resp : HResponse, 
                             chunked : Boolean, 
                             stream : HStream, 
                             session : Option[HSess]) 
  extends HOut(aliveReq, resp, chunked, stream, session) {

  final protected object bos extends java.io.ByteArrayOutputStream {
    def getBuf : Array[Byte] = buf
    def getLength : Int = count
  }
  
  private var closed = false // single thread
    
  final def isHeaded : Boolean = false
  final def isClosed : Boolean = closed
    
  def write(ar : Array[Byte], offset : Int, length : Int) : Unit = bos.write(ar, offset, length)
  
  final def close : Unit = if(!closed) {
    closed = true
    closeMe
    
    // 1. set and write header
    if(!aliveReq) resp.setHeader("Connection", "Close") // else - done in HConn
    setContentEncoding
    
    if(chunked) {
      resp.removeHeader("Content-Length")
      resp.setHeader("Transfer-Encoding", "chunked")
    }
    else resp.setHeader("Content-Length", bos.getLength.toString)

    writeSessionCookie
    stream.out.write(resp.getBytes)
    
    // 2. write content and real close
    if(chunked) {
      stream.out.write((bos.getLength.toHexString.toLowerCase + "\r\n").getBytes("ISO-8859-1"))
      stream.out.write(bos.getBuf, 0, bos.getLength)
      stream.out.write("\r\n".getBytes("ISO-8859-1"))
      stream.out.write("0\r\n\r\n".getBytes("ISO-8859-1"))
    }
    else stream.out.write(bos.getBuf, 0, bos.getLength)   
    
    restamp
    closeStream
  }
  
  protected def setContentEncoding : Unit = resp.removeHeader("Content-Encoding") // gzip is supported only
  protected def closeMe = bos.close
}
  
  
final protected class HGzipOut(aliveReq : Boolean, 
                               resp : HResponse, 
                               chunked : Boolean, 
                               stream : HStream, 
                               session : Option[HSess]) 
  extends HBufferedOut(aliveReq, resp, chunked, stream, session) {

  private val zos = new java.util.zip.GZIPOutputStream(bos)
  override def write(ar : Array[Byte], offset : Int, length : Int) : Unit = zos.write(ar, offset, length)
  override protected def setContentEncoding = resp.setHeader("Content-Encoding", "gzip")
  override protected def closeMe = zos.close
}

