package zgs.httpd


trait HServer {

  //------------------ to implement -----------------------------------
  
  protected def apps : Seq[HApp]
  
  protected def ports : Seq[Int]
  
  protected def talkPoolSize : Int  // threads
  protected def talkQueueSize : Int // tasks
  
  protected def selectorPoolSize : Int // threads
  
  //------------------------- to override ------------------------------
  
  protected def name = "tiscaf" // ti-ny Sca-la f-ramework
  
  protected def stopPort : Int = 8911
  
  protected def tcpNoDelay : Boolean = false // use 'true' for benchmarking
  
  protected def readBufSize : Int  = 512  // nio read...
  protected def writeBufSize : Int = 4096 // ... and write buffers sizes

  protected def connectionTimeoutSeconds : Int  = 30  // is used for Keep-Alive mode (if turned on)
                                                      // *and* in multiplexer to dispose dead channels
  
  protected def interruptTimeoutMillis : Int = 1000 // for stopping: take HLets a chance to finish 

  protected def onError(e : Throwable) : Unit = e.printStackTrace

  // override if you want more elaborated shutdown procedure (and replace zgs.httpd.HStop)
  protected def startStopListener : Unit = scala.concurrent.ops.spawn {
    val ss = new java.net.ServerSocket(stopPort)
    println(name + " stop listener is listening to port " + stopPort)
    val s  = ss.accept
    val ar = new Array[Byte](256)
    s.getInputStream.read(ar)
    if(new String(ar, "ISO-8859-1") startsWith("stop")) { s.close; ss.close; stop }
    else error(name + ": invalid stop sequence")
  }

  //-------------------------- user API ---------------------------------
  
  final def start : Unit = synchronized {
    talksExe = new zgs.sync.SyncExe(talkPoolSize, talkQueueSize, "Talks")
    plexer = createPlexer
    plexer.start
    ports.foreach { port => plexer.addServer(peerFactory, port) }
    startStopListener
    stopped.set(false)
    println(name + " server was started on port(s) " + ports.mkString(", "))
  }
  
  final def stop : Unit = synchronized {
    if(!stopped.get) {
      stopped.set(true)
      talksExe.shutdown
      Thread.sleep(interruptTimeoutMillis)
      talksExe.shutdownNow 
      plexer.stop
      println(name + " server stopped")
    } 
  }
  
  //--------------------------- internals -------------------------------
  
  // we need to share states between 'start' and 'stop'.
  // nothing must be started in init
  private var talksExe : zgs.sync.SyncExe = _
  private var plexer : HPlexer            = _
  
  
  private val connTimeoutMillis : Long = connectionTimeoutSeconds * 1000
  
  private val stopped = new zgs.sync.SyncBool(true)
  
  private def createPlexer = new HPlexer {
    protected def connectionTimeout : Long      = HServer.this.connTimeoutMillis
    protected def selectorPoolSize : Int        = HServer.this.selectorPoolSize
    protected def tcpNoDelay : Boolean          = HServer.this.tcpNoDelay
    protected def onError(e : Throwable) : Unit = HServer.this.onError(e)
  }
  
  import java.nio.channels. { SelectionKey, SocketChannel }
  
  // key place
  private def peerFactory(key : SelectionKey) : HPeer = new HPeer { self =>
    
    def readBufSize : Int                         = HServer.this.readBufSize
    def writeBufSize : Int                        = HServer.this.writeBufSize
    def queueWakeup(want : PeerWant.Value) : Unit = plexer.toWake(key, want)
    protected def channel : SocketChannel         = key.channel.asInstanceOf[SocketChannel]
    protected def onError(e : Throwable) : Unit   = HServer.this.onError(e)
    protected def connectionTimeout : Long        = HServer.this.connTimeoutMillis
    
    private val connection = new HConn(new HStream(self), apps, HServer.this.connectionTimeoutSeconds)

    def submit : Unit = if(!stopped.get) 
      talksExe.submit(new Runnable { def run { try { connection.process } catch { case e => onError(e) } } })
  }
  
}
