package zgs.httpd

import scala.collection. { mutable => mute }


final protected class HSess(app : HApp, conn : HConn, resp : HResponse) {
  
  //---------------------- delegating here from HTalk user API ---------

  def sid : String                          = session._1 
  def apply(key : Any) : Option[Any]        = session._2.get(key)
  def update(key : Any, value : Any) : Unit = session._2(key) = value
  def remove(key : Any) : Option[Any]       = { val v = session._2.get(key); session._2 - key; v } 
  def clear : Unit                          = session._2.clear
  def keys : Seq[Any]                       = session._2.keys.toList
  
  //----------------------- SPI ---------------------
  
  def writeSessionCookie : Unit = {
    def writeIt = resp.setHeader("Set-Cookie", HApp.cookieKey + "=" + session._1)
    app.tracking match {
      case HTracking.UriAndCookie => writeIt
      case HTracking.CookieOnly   => writeIt
      case _                      =>
    }
  }  

  def restamp    = HSess.restamp(app, session._1)
  def isValid    = HSess.isValid(app, session._1)
  def invalidate = HSess.invalidate(app, session._1)

  //-------------------- internals ---------------------------------------------
  
  private lazy val session : (String, scala.collection.mutable.Map[Any,Any]) = extractSid match {
    case None    => HSess.create(app)
    case Some(x) => HSess.bag(app, x) match { // extracted sid may invalidated already
      case None      => HSess.create(app)
      case Some(bag) => 
        //println("found: " + bag.keySet); 
        (x, bag)
    }
  }
  
 // uri has higher priority rather cookie
  private def extractSid : Option[String] = app.tracking match {
    case HTracking.UriOnly      => extractSidFromUri
    case HTracking.CookieOnly   => extractSidFromCookie
    case HTracking.UriAndCookie => extractSidFromUri match {
      case None => extractSidFromCookie
      case x    => x
    }
    case _ => None
  }
  
  private def extractSidFromUri : Option[String] = conn.headerData.uriExt match {
    case Some(x) => 
      val parts = x.split("=", 2)
      if(parts.size == 2 && parts(0) == HApp.sidKey && HSess.sidIsValid(parts(1))) Some(parts(1))
      else None
    case _ => None
  }
  
  private def extractSidFromCookie : Option[String] = conn.headerData.header("cookie") match {
    case None    => None
    case Some(x) => 
      val sids = for {
        pair <- x.split(";", 2)
        val parts = pair.split("=", 2)
        if(parts.size == 2) && parts(0).trim == HApp.cookieKey && HSess.sidIsValid(parts(1).trim)
      } 
      yield parts(1).trim
          
      if(sids.size == 1) Some(sids(0)) else None
  }
  
}


private object HSess {
  
  def create(app : HApp) : (String,mute.Map[Any,Any]) = 
    withApp(app, _.create, { val aS = new HAppSess; bags(app) = aS; aS.create }) 
    
  def bag(app : HApp, sid : String) : Option[mute.Map[Any,Any]] = withApp(app, _.bag(sid), None) 
  def restamp(app : HApp, sid : String) : Unit                  = withApp(app, _.restamp(sid), None)
  def isValid(app : HApp, sid : String) : Boolean               = withApp(app, _.isValid(sid), false)
  def invalidate(app : HApp, sid : String) : Unit               = withApp(app, _.invalidate(sid), {})
  
  //----------------- internals --------------------
  
  private val alpha = "abcdefghijklmnopqrstuvwxyz"
  private val symbols = alpha + alpha.toUpperCase + "0123456789"
  private val symLength = symbols.length
  private val idLength = 16
  private val random = new java.security.SecureRandom
  
  private def withApp[T](app: HApp, some : HAppSess => T, none : => T) : T = 
    bags.synchronized { bags.get(app) match { case None => none; case Some(aS) => some(aS) }}
  
  private def newSid : String = synchronized {
    val buf = new StringBuilder
    (1 to idLength).foreach(_ => buf.append(symbols.charAt(random.nextInt(symLength))))
    buf.toString
  }
  
  private def sidIsValid(sid : String) : Boolean = 
    if(idLength != sid.length) false 
    else (0 until idLength).find(i => symbols.indexOf(sid.charAt(i)) < 0).isEmpty
  
  private def sidKey = "sid"
  private def cookieKey = "TISCAF_SESSIONID"
  
  private val bags = new mute.HashMap[HApp,HAppSess] // hand-written synchronization
    
  private val millisInMin = 60000 
  private def cleanExpired = bags.synchronized { 
    bags.keys.foreach(app => bags(app).cleanExpired(app.sessionTimeoutMinutes * millisInMin)) 
  }
  
  new java.util.Timer(true) // daemon
    .scheduleAtFixedRate(
      new java.util.TimerTask { def run { cleanExpired } },
      15000, 15000 // every 15 seconds 
    )
  
  //--------------------------------------------
  
  private class HAppSess {
    
    type StampBag = (Long,mute.Map[Any,Any])
    
    def create : (String,mute.Map[Any,Any]) = {
      val sidMap = new mute.HashMap[Any,Any] with mute.SynchronizedMap[Any,Any]
      val sid = newSid
      appBags.synchronized {
        appBags(sid) = (now, sidMap)
        (sid, sidMap)
      }
    }
  
    def bag(sid : String) : Option[mute.Map[Any,Any]] = withApp(sid, { bag => Some(bag._2) }, None)
      
    def restamp(sid : String) : Unit    = withApp(sid, { bag => appBags(sid) = (now, bag._2) }, {})
    def isValid(sid : String) : Boolean = withApp(sid, _ => true, false)
    def invalidate(sid : String) : Unit = appBags.synchronized { appBags - sid }
    
    def cleanExpired(timeout : Int) : Unit = appBags.synchronized {
      val timeX = now - timeout
      appBags.keys.filter(sid => appBags(sid)._1 < timeX).foreach(sid =>  appBags - sid)
    }
    
    //--------------
    
    private val appBags : mute.Map[String,(Long,mute.Map[Any,Any])] = new mute.HashMap[String,(Long,mute.Map[Any,Any])]
    private def now = System.currentTimeMillis
    
    private def withApp[T](sid: String, some : StampBag => T, none : => T) : T = 
      appBags.synchronized { appBags.get(sid) match { case None => none; case Some(tb) => some(tb) }}
  }
  
}
