package zgs.httpd


final class HTalk(private val app : HApp, 
                  private val connection : HConn, 
                  private val stream : HStream,
                  private val paramsEncoding : String) {
  
  //----------------------------------------------------------------
  //---------------------------- main user API ---------------------
  //----------------------------------------------------------------
  
  //------------------ request info -----------

  object req {
    // common
    def method : HReqType.Value  = connection.headerData.reqType
    def host : Option[String]    = connection.headerData.host
    def port : Option[String]    = connection.headerData.port
    def uriPath : String         = connection.headerData.uriPath
    def uriExt : Option[String]  = connection.headerData.uriExt
    def query : String           = connection.headerData.query
    def remoteIp : String        = stream.remoteIp 
  
    // header
    def header(key : String): Option[String]      = connection.headerData.header(key)
    def headerKeys : scala.collection.Set[String] = connection.headerData.headerKeys
  
    // parameters
    def paramKeys : Seq[String]              = connection.paramKeys.map(decode _).toSeq
    def params(key : String) : Seq[String]   = connection.params(encode(key)).map(decode _)
    def param(key : String) : Option[String] = connection.param(encode(key)).map(decode _)
    
    // POST/application/octet-stream case
    def octets : Option[Array[Byte]] = connection.octets
  }

  // ------------- setting status / response header -----------

  def setStatus(code : HStatus.Value) : HTalk               = { checkHead; resp.setStatus(code); this }
  def setStatus(code : HStatus.Value, msg : String) : HTalk = { checkHead; resp.setStatus(code, msg); this }
  
  def setHeader(name : String, value : String) : HTalk = { checkHead; resp.setHeader(name, value); this }
  def removeHeader(name : String) : Option[String]     = { checkHead; resp.removeHeader(name) }
  def getHeader(name : String) : Option[String]        = { checkHead; resp.getHeader(name) }
  
  def setContentLength(length : Long) : HTalk        = setHeader("Content-Length", length.toString)
  def setCharacterEncoding(charset : String) : HTalk = setHeader("Character-Encoding", charset)
  def setContentType(cType : String) : HTalk         = setHeader("Content-Type", cType)
  
  //------------------- output/close -------------------------
  
  def write(ar : Array[Byte], offset : Int, length : Int) : HTalk = { 
    getOut.write(ar, offset, length) 
    this
  }
  
  def write(ar : Array[Byte]) : HTalk = write(ar, 0, ar.length)
  
  def close : Unit = getOut.close // it is safe to call multiple close
  
  //------------------------ session-related -------------------
  
  object ses {
    def tracking : HTracking.Value            = app.tracking
    def isAllowed : Boolean                   = tracking != HTracking.NotAllowed
    def isValid : Boolean                     = isAllowed && session.get.isValid
    def clear : Unit                          = if(isAllowed) session.get.clear
    def invalidate : Unit                     = if(isAllowed) session.get.invalidate
    
    def idKey : String    = HApp.sidKey
    def id : String       = session.get.sid
    def idPhrase : String = (new StringBuilder).append(";").append(idKey).append("=").append(id).toString
  
    def apply(key : Any) : Option[Any]        = session.get(key)
    def update(key : Any, value : Any) : Unit = session.get(key) = value
    def keys : Seq[Any]                       = session.get.keys
    def remove(key : Any) : Option[Any]       = session.get.remove(key)
  }
  
  //-- also may be useful for API user
  
  def isClosed : Boolean = if(out.isEmpty) false else out.get.isClosed
  
  //---------------------- internals -----------------------------
  
  private val resp    = new HResponse
  private val session : Option[HSess] = 
    if(app.tracking == HTracking.NotAllowed) None
    else Some(new HSess(app, connection, resp))
  
  // we can not define 'out' before header's Content-Type is defined (if it will)
  private var out : Option[HOut] = None
  
  private def checkHead = if(out.isDefined && out.get.isHeaded) {
    stream.dispose
    throw new HError("header is already sent")
  }
    
  private def getOut : HOut = { 
    if(out.isEmpty) out = {
      def aliveReq = try { connection.headerData.isPersistent } catch { case _ => false }
      
      if(gzipable)          Some(new HGzipOut(aliveReq, resp, app.chunked, stream, session))
      else if(app.buffered) Some(new HBufferedOut(aliveReq, resp, app.chunked, stream, session))
      else                  Some(new HIdentOut(aliveReq, resp, app.chunked, stream, session))
    }
    out.get
  }

  private def gzipable = {
    def accepted = req.header("Accept-Encoding") match {
      case None    => false
      case Some(x) => x.contains("gzip")
    }
    
    def mimeOk = resp.getHeader("Content-Type") match {
      case None    => false
      case Some(x) => HMime.gzipable.contains(x.split(";")(0).trim)
    }
    
    app.gzip && accepted && mimeOk
  }
  
  private def encode(s : String) = try { java.net.URLEncoder.encode(s, paramsEncoding) } catch { case _ => s }
  private def decode(s : String) = try { java.net.URLDecoder.decode(s, paramsEncoding) } catch { case _ => s }
}

